;; Generated package description from tex-parens.el  -*- no-byte-compile: t -*-
(define-package "tex-parens" "0.3" "like lisp.el but for tex" '((emacs "27.1") (auctex "14.0.5")) :commit "ae2a66e3bfd1cdc37cc3cb2416d940aca1460c8b" :authors '(("Paul D. Nelson" . "nelson.paul.david@gmail.com")) :maintainer '("Paul D. Nelson" . "nelson.paul.david@gmail.com") :keywords '("tex" "convenience") :url "https://github.com/ultronozm/tex-parens.el")
